import plotly.graph_objects as go

def getShapeMatrix(ro=1.3, N=1000, gpu=False):
    if gpu:
        import cupy as np
    else:
        import numpy as np

    def f(ro, N):
        ''' give the surface map matrix '''
        # polar coordinate
        theta = np.linspace(0, 2 * np.pi, N)
        x = ro * np.cos(theta)
        y = ro * np.sin(theta)

        u, v = np.meshgrid(x, y)
        RO = u ** 2 + v ** 2
        # map z to u,v
        Z = 1 - u / RO - RO
        return x, y, Z

    x, y, Z = f(ro, N)
    # remap z to the grid index
    index_x, index_y = np.argsort(x), np.argsort(y)
    Y, X = np.meshgrid(index_x, index_y)  # meshgrid always will use numpy array NOT mxnet NDArray
    Z = Z[X, Y]
    return Z

def love():
    res = getShapeMatrix(ro=1.3, N=150, gpu=False)[12:129, :100]
    # res[res<-0.9] = 0
    res[res > 5] = 0
    res = 1 - res
    res[res > 1] = 1
    res[res < -3] = -3
    res = res + 3
    res /= 4

    fig = go.Figure(data=[go.Surface(z=res,
                                     # colorscale='Viridis',
                                     showscale=False,
                                     opacity=0.9)])
    fig.update_layout(title='noise distribution', autosize=False,
                      width=700, height=700,
                      )
    fig.update_traces(contours_z=dict(show=True, usecolormap=True,
                                      highlightcolor="limegreen", project_z=True))
    fig.show()

if __name__ == "__main__":
    love()




import numpy as np
from skimage.transform import hough_circle, hough_circle_peaks
from skimage.feature import canny
import cv2
import matplotlib.pyplot as plt


def circle_detect(image, rangeR=(200, 250), total_num_peaks=1):
    """
    :param rangeR: detected circles' Radius range
    :param total_num_peaks: detect circles number
    return (accums, cx, cy, radii)"""
    # image = cv2.bilateralFilter(image, 21, 75, 75)
    image = cv2.medianBlur(image,15)
    edges = canny(image,  low_threshold=3, high_threshold=10) # sigma=3,
    # plt.imshow(edges)
    # plt.show()
    hough_radii = np.arange(rangeR[0], rangeR[1], 2)
    hough_res = hough_circle(edges, hough_radii)

    accums, cx, cy, radii = hough_circle_peaks(hough_res, hough_radii,
                                               total_num_peaks=total_num_peaks)
    return accums, cx, cy, radii


if __name__ == "__main__":
    from guang.cv.video import getFrame
    import matplotlib.pyplot as plt
    from skimage.draw import circle_perimeter
    from skimage import color
    
    # dst = cv2.pyrMeanShiftFiltering(image, 10, 100)   #边缘保留滤波EPF
    fps, size_17, img = getFrame(filename=r'C:\Users\beidongjiedeguang\Desktop\实验\62.avi',
                                       frameNum=723,
                                 )
    image = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    accums, cx, cy, radii = circle_detect(image, rangeR=[100, 250], total_num_peaks=1)
    image = color.gray2rgb(image)
    for center_y, center_x, radius in zip(cy, cx, radii):
        circy, circx = circle_perimeter(center_y, center_x, radius)
        image[circy, circx] = (220, 20, 20)

    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(20, 10))
    ax.imshow(image, cmap=plt.cm.gray)
    plt.show()



import cv2
from pyprobar import bar
import matplotlib.pyplot as plt

def getFrame(filename, frameNum=10,  W=(0, -1), H=(0, -1), gray=False):
    """ Get a specific frame in the video
    :param W: ROI W range
    :param H: ROI H range
    :return : (TotolFrameNum, fps, size), ROI
    """
    cap = cv2.VideoCapture(filename)
    fps = cap.get(cv2.CAP_PROP_FPS)
    size = (int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))  # 获取视频尺寸
    TotolFrameNum = int(cap.get(7))
    cap.set(cv2.CAP_PROP_POS_FRAMES, frameNum)  # 设置要获取的帧号
    ret, frame = cap.read()
    if not gray:
        ROI = frame[H[0]:H[1], W[0]:W[1], :]
    else:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        ROI = frame[H[0]:H[1], W[0]:W[1]]

    cap.release()
    return (TotolFrameNum, fps, size), ROI

def resample(inputvideo, outputvideo, W=(0,-1), H=(0,-1), Time=None, samplerate=1, fps=None):
    """重采样，可指定视频对应采样率或fps、ROI(采样区域)、采样时长
    :param W: (宽度开始位置，宽度结束位置)
    :param H: (高度开始位置，高度结束位置)
    :param Time: 采样时长
    :param samplerate: 采样率，如设置0.1则表示对原视频每10帧采样一次
    :param fps: 如果设置了samplerate就不要再设置fps了, 它们二选一
    """
    cap = cv2.VideoCapture(inputvideo)
    FPS = cap.get(cv2.CAP_PROP_FPS)

    SIZE = [int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))]  # 获取视频尺寸

    fourcc = cv2.VideoWriter_fourcc(*'XVID')  # 编码格式

    TotolFrameNum = int(cap.get(7))  # == cv2.CV_CAP_PROP_FRAME_COUNT
    if Time is None:
        FrameNum = TotolFrameNum
    else:
        FrameNum = int(FPS * Time)
    print(f"原视频总帧：{TotolFrameNum}")
    if W == (0, -1):
        W = (0, int(SIZE[0]))
    else:
        SIZE[0] = W[1] - W[0]
    if H == (0, -1):
        H = (0, int(SIZE[1]))
    else:
        SIZE[1] = H[1] - H[0]

    step = round(1 / samplerate)
    if fps is None:
        fps = round(FPS * samplerate)
    else:
        step = round(FPS/fps)
    video = cv2.VideoWriter(outputvideo, fourcc, fps, tuple(SIZE))
    count = 0
    for idx in range(FrameNum):
        bar(idx, FrameNum)
        if idx % step != 0: continue
        cap.set(cv2.CAP_PROP_POS_FRAMES, idx) # 为了均匀采样
        ret, frame = cap.read()
        ROI = frame[H[0]:H[1], W[0]:W[1], :]
        if not ret:
            break
        video.write(ROI)
        count += 1
    cap.release()
    print(f"总帧数：{count}")

def embedFrameInfo(inputvideo, outputvideo, fps=None):
    cap = cv2.VideoCapture(inputvideo)
    if fps is None:
        fps = cap.get(cv2.CAP_PROP_FPS)

    size = (int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))  # 获取视频尺寸

    fourcc = cv2.VideoWriter_fourcc(*'XVID')  # 编码格式

    video = cv2.VideoWriter(outputvideo, fourcc, fps, size)

    TotolFrameNum = int(cap.get(7)) # == cv2.CV_CAP_PROP_FRAME_COUNT
    for idx in range(TotolFrameNum):
        ret, frame = cap.read()
        if not ret:
            break
        cv2.putText(frame, f'current frame: {idx}', (50, 60), cv2.FONT_HERSHEY_PLAIN, 1.0, (0, 255, 0), 2, 1)
        cv2.putText(frame, f'time: {idx/ fps:.2f} S', (50, 100), cv2.FONT_HERSHEY_PLAIN, 1.0, (0, 255, 0), 2, 1)
        video.write(frame)
        bar(idx, TotolFrameNum)
        # cv2.imshow("zd1", frame)
        # if cv2.waitKey(int(10/fps)) & 0xFF == ord('q'): # 0xFF的使用：https://blog.csdn.net/i6223671/article/details/88924481
        #     # 实际上 任何一个整数与0xff做与运算，将会取得这个数最后八位
        #     break

    cap.release()
    # if show:
    #     cv2.destroyAllWindows()



import numpy as np
import scipy.signal as signal
from guang.sci.fft import space2fre_2d


def culc_frequency(x, y):
    idx_max = signal.argrelextrema(y, np.greater)[0]
    idx_min = signal.argrelextrema(y, np.less)[0]
    Lmax, Lmin = len(idx_max), len(idx_min)
    if Lmax<=1 and Lmin<=1:
        raise ValueError("无振荡")
    
    if Lmax > Lmin:
        N = Lmax-1
        frequency = N/(x[idx_max[-1]] - x[idx_max[0]])
    else:
        N = Lmin - 1
        frequency = N / (x[idx_min[-1]] - x[idx_min[0]])
    return frequency, idx_max, idx_min


def culc_fig_fre(mat, method, dw=1,  dh=1, fre_min=1., FRE_MAX=15, neighb=2):
    """
    :param mat: 2d array
    :param method: options "1d" or "2d"
    """
    Nh, Nw = mat.shape
    fre_max = 1 / (2 * dw)

    def get_fre(N):
        delta_fre = 2 * fre_max / N
        if N % 2 == 0:
            fre = np.linspace(-fre_max, fre_max, N) - delta_fre / 2
        else:
            fre = np.linspace(-fre_max, fre_max, N)
        idx = np.argwhere(fre >= 0)
        return idx, fre[idx]

    def get_fft_1d():
        # idx_h, fre_h = get_fre(dh, Nh)
        idx_w, fre_w = get_fre(Nw)
        IDX = np.argwhere(fre_w > FRE_MAX)[0][0]
        F = np.fft.fft2(mat)
        F[:, IDX:-IDX] = 1e-14 + 0j  # F[IDX:-IDX, :]
        Fu = F[0, :]  # F[:,0]
        x = np.arange(0, dw * Nw, dw)
        return x, Fu

    if method == "1d":
        x, Fu = get_fft_1d()
        res = np.real(np.fft.ifft(Fu))
        # x = np.linspace(0, dw*Nw, Nw)
        # fig = plt.figure()
        # plt.plot(x, res)
        # plt.show()
        frequncy = culc_frequency(x, res)[0]

    elif method == "w_1d":
        _, Fu = get_fft_1d()
        Fu_shift = np.fft.fftshift(Fu)
        idx_x, x = get_fre(Nw)
        F = Fu_shift[idx_x]
        x = x.reshape(-1, )

        amplitude = np.real(np.log(F)).reshape(-1,)

        idx_1 = np.argwhere(x > fre_min)
        idx_2 = np.argmax(amplitude[idx_1])

        idx = int(idx_1[0] + idx_2)

        def weight_fre(fre, weight, idx, neighb = 2):
            IDX = np.arange(idx-neighb, idx+neighb+1, 1)
            res = 0
            for i in IDX:
                res += fre[i] * weight[i]
            return res/np.sum(weight[IDX])

        frequncy =  weight_fre(x, amplitude, idx, neighb=neighb)# x[idx]


    elif method == "2d":
        x, y, amplitude = space2fre_2d(mat, dh, dw)
        argmat = x > fre_min
        frequncy = x[argmat][np.argmax(amplitude[argmat])]
    else:
        raise ValueError("bad method")
    return frequncy


if __name__ == "__main__":
    from guang.cv.video import getFrame
    import matplotlib.pyplot as plt
    INFO, mat = getFrame(r'C:\Users\beidongjiedeguang\Desktop\实验\60degree\1_resample.avi', 1400,
                         W=(52, 618), H=(43, 580),
                         gray=1)
    fig2 =plt.figure()
    plt.imshow(mat)
    plt.show()

    dtheta = 3
    dx = dtheta / (618 - 52)
    print(culc_fig_fre(mat, method=1, dw=dx, FRE_MAX=13))
    print(culc_fig_fre(mat, method=2, dw=dx, fre_min=3))

    


import numpy as np
import cupy as cp
from numpy.fft import fft,fft2, fftshift
# from guang.sci.utils import culc_frequency

def fft_conv1d(a,b):
    n = len(a)+len(b)-1
    N = 2**(int(np.log2(n))+1)
    A = np.fft.fft(a, N)
    B = np.fft.fft(b, N)
    return np.fft.ifft(A*B)[:n]

def fft_conv1d_cupy(a,b):
	
    n = len(a)+len(b)-1
    N = 2**(int(cp.log2(n))+1)
    A = cp.fft.fft(a, N)
    B = cp.fft.fft(b, N)
    return cp.fft.ifft(A*B)[:n]

def fft_conv2d(a, b):

    ha, wa = a.shape
    hb, wb = b.shape
    w = wa + wb - 1
    h = ha + hb - 1
    Nw = 2**(int(np.log2(w))+1)
    Nh = 2**(int(np.log2(h))+1)
    A = np.fft.fft2(a, (Nh, Nw))
    B = np.fft.fft2(b, (Nh, Nw))
    return np.fft.ifft2(A*B)[:h, :w]

def fft_conv2d_cupy(a, b):
	
    ha, wa = a.shape
    hb, wb = b.shape
    w = wa + wb - 1
    h = ha + hb - 1
    Nw = 2**(int(np.log2(w))+1)
    Nh = 2**(int(np.log2(h))+1)
    A = cp.fft.fft2(a, (Nh, Nw))
    B = cp.fft.fft2(b, (Nh, Nw))
    return cp.fft.ifft2(A*B)[:h, :w]


def space2fre(x, y):
    """ Transform space domain (x, y) to the frequency domain (frequency, amplitude)
    """
    x, y = x.reshape(-1, 1)[...,-1], y.reshape(-1, 1)[...,-1]
    x_max = x.max()
    x_min = x.min()
    N = len(x)
    delta_x = (x_max - x_min) / (N - 1)
    f_max = 1 / (2 * delta_x)
    delta_f = 2 * f_max / (N - 1)
    if N%2==0:
        fre_x = np.linspace(-f_max, f_max, N)-delta_f/2
    else:
        fre_x = np.linspace(-f_max, f_max, N)
    fre_y_shift = fftshift(fft(y))
    magnitude_energy = np.abs(fre_y_shift) / (N - 1)
    amplitude_y = magnitude_energy * 2 #  则表示原来振幅
    idx = np.argwhere(fre_x>=0)
    return fre_x[idx].reshape(-1, 1), amplitude_y[idx].reshape(-1, 1)

def space2fre_2d(mat, dh=1, dw=1):
    """mat: 2d numpy array"""

    Nh, Nw = mat.shape
    def get_fre(dx, N):
        fre_max = 1 / (2 * dx)
        # print("fre_max", fre_max)
        delta_fre = 2 * fre_max / N
        if N % 2 == 0:
            fre = np.linspace(-fre_max, fre_max, N) - delta_fre / 2
        else:
            fre = np.linspace(-fre_max, fre_max, N)
        idx = np.argwhere(fre >= 0)
        return idx, fre[idx]
    idx_h, fre_h = get_fre(dh, Nh)
    idx_w, fre_w = get_fre(dw, Nw)
    x, y = np.meshgrid(fre_w, fre_h)
    F = fftshift(fft2(mat))
    magnitude_energy = np.abs(F) / ((Nh - 1) * (Nw - 1))
    amplitude = magnitude_energy * 2
    return x, y, amplitude[idx_h][:, -1, :][:, idx_w][:, :, -1]




from guang.cv.MajorTone import *
import cv2
import matplotlib.pyplot as plt
from skimage import data
from skimage.transform import match_histograms

def histogramMatching():
    reference = data.coffee()
    image = data.chelsea()

    matched = match_histograms(image, reference, multichannel=True)

    fig, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, figsize=(8, 3),
                                        sharex=True, sharey=True)
    for aa in (ax1, ax2, ax3):
        aa.set_axis_off()

    ax1.imshow(image)
    ax1.set_title('Source')
    ax2.imshow(reference)
    ax2.set_title('Reference')
    ax3.imshow(matched)
    ax3.set_title('Matched')

    plt.tight_layout()
    plt.show()

def generateBackGround(img,show_plot=1):
    '''
    input: BGR image array
    return: a background image ,has the same shape with input image.
    '''
    img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h,w,_ = img.shape
    feature = cv2.resize(img,dsize=(40,40))
    color1 = MajorTones.getCenters(feature, n_cls = 2, m_cls=1)[0][0]
#     color2 = MajorTones.getCentroid(feature, n_cls = 2)

    color = color1.astype('uint8')
#     color = color2.astype('uint8')

    if show_plot:
        plt.figure()
        plt.subplot(121)
        plt.imshow(cv2.cvtColor(img,cv2.COLOR_HSV2RGB))
        plt.subplot(122)
        plt.imshow(cv2.cvtColor(color.reshape(1,1,3),cv2.COLOR_HSV2RGB).reshape(1,1,3))
        plt.show()
    H,W = 55,55
    gener_backg = np.ones((H,W,3)).astype('uint8')
    gener_backg[:,:,0] = gener_backg[:,:,0]*color[0] # H
    gener_backg[:,:,1] = gener_backg[:,:,1]*color[1] # S
    gener_backg[:,:,2] = gener_backg[:,:,2]+250      # V

    
    if show_plot:
        plt.figure()
        plt.imshow(cv2.cvtColor(gener_backg,cv2.COLOR_HSV2RGB))
        plt.show()
    
    i = 1
    c_max = gener_backg[i].max()
    R2 = H**2+W**2
    for x in range(W):
        for y in range(H):
            r2 = (x-W/2)**2 + (y-H/2)**2 # 圆心转移到中心
            c = c_max * r2//R2
            gener_backg[y,x,i] = c
    
    gener_backg=cv2.cvtColor(gener_backg,cv2.COLOR_HSV2RGB)
    gener_backg=cv2.resize(gener_backg,(h,w),interpolation=cv2.INTER_NEAREST)
    if show_plot:
        plt.figure()
        plt.imshow(gener_backg)
        plt.show()

    return gener_backg

if __name__ == "__main__":
    def test_generateBackGround():
        img = cv2.imread("pic/hue1.png")
        generateBackGround(img, show_plot=1)

    # test_generateBackGround()
    histogramMatching()


# import sys,os
# sys.path.append(os.getcwd())
from guang.Utils.toolsFunc import sort_count

import numpy as np
import matplotlib.pyplot as plt

from sklearn.cluster import KMeans
# from sklearn.cluster import k_means 
from PIL import Image
import colorsys

import cv2



class MajorTones:
    
    @staticmethod
    def preProces(imgName):
        ''' 简单预处理下'''
        feature = Image.open(imgName)
        feature = feature.convert("RGB")
        try:
            # resize一下,提高速度
            feature = feature.resize((150,150))
            feature = np.array(feature)
            # 因为有的图片色彩通道是4个
            feature = feature[:,:,:3]

            if feature.dtype == 'uint8':
                feature = feature/255.
        except:
            print(f'打开图片失败, 跳过该图片：{imgName}')
            return []

        return feature


    
    @staticmethod
    def plot_simgle_sample(img_path, n_cls=2):
        '''测试k_means效果'''
        km = KMeans(n_clusters=n_cls)
        plt.subplot(1,3,1)
        plt.imshow(Image.open(img_path))
        # 主基调图片
        keynote  = MajorTones.preProces(img_path)
        plt.subplot(1,3,2)
        plt.imshow(keynote)
        h,w,channel = keynote.shape
        keynote = km.fit(keynote.reshape(h*w, channel))

        plt.subplot(1,3,3)
        plt.imshow(keynote.cluster_centers_.reshape(n_cls,1,3))
        plt.show()
        print(keynote.cluster_centers_)

    @staticmethod
    def getCenters(feature, n_cls, m_cls):
        '''获得多个色彩聚类中心, n_cls = kmeans中的k个聚类
        m_cls： 它接受两中类型的值：
        如果它的值是整数，则表示返回k个聚类中的前m个聚类
        如果它的值是0~1之间的浮点数，如0.5,则它会返回前百分之五十的聚类结果
        output:
        key_centers: 返回前m个类的中心坐标
        key_percent: 返回的key_centers中各个元素所占的比例
        '''
        import warnings
        warnings.filterwarnings('ignore')
        (h, w, channel) = feature.shape
        feature = feature.reshape(h * w, channel)
        res = KMeans(n_clusters=n_cls).fit(feature)
        predict = res.predict(feature)
        cluster_centers = res.cluster_centers_
        predict = sort_count(predict)# [(2,5),(cluster, counts)...]
        
        # 可能聚类结果没有达到n_cls设定值，这里重新赋值
        n_cls = len(predict)
        pred_idx, counts = [predict[i][0] for i in range(n_cls)], [predict[i][1] for i in range(n_cls)]
        Totle = sum(counts)
        # print(predict)
        if type(m_cls) == int:
            # 这里m_cls也是一样
            if m_cls > n_cls:
                m_cls = n_cls

            key_predict = predict[:m_cls]
            key_idx, _ = [key_predict[i][0] for i in range(m_cls)], [key_predict[i][1] for i in range(m_cls)]


        elif 0 < m_cls < 1:
            S = 0
            for i, _ in enumerate(counts):
                S += _
                if S / Totle >= m_cls:
                    break
            key_idx = pred_idx[:i + 1]
        else:
            raise ' '
        # print(key_idx)
        key_centers = cluster_centers[key_idx]
        count_percent = np.array(counts) / Totle
        key_percent = count_percent[:len(key_idx)]  # key_percent储存返回的类所占的比例
        return key_centers, key_percent
        
    @staticmethod
    def getCentroid(feature, n_cls = 2):
        '''获取主色调的RGB值
        inputParam: 
        feature: 3D rgb array
        n_cls: Please keep it unchanged
        '''
        (h, w, channel) = feature.shape
        feature = feature.reshape(h*w, channel)
        res = KMeans(n_clusters=n_cls).fit(feature)
        L = len(res.labels_)

        #　选取较多的那个簇
        if L - np.count_nonzero(res.labels_) > L/2: flag = 0 
        else: flag = 1 
        feature = feature[res.labels_==flag]   
        # 进行第二次聚类
        res = KMeans(n_clusters=n_cls).fit(feature)
        L = len(res.labels_)
        if L - np.count_nonzero(res.labels_) > L/2: flag = 0
        else: flag = 1

        return res.cluster_centers_[flag]
    
    
    @staticmethod
    def get_dominant_color(image):
        '''从网上down来的,运算量较小,可能适用于某些场景,
        至于适用于哪些,我咋知道
        '''
        image = Image.open(image)
        #颜色模式转换，以便输出rgb颜色值
        image = image.convert('RGBA')
        #生成缩略图，减少计算量，减小cpu压力
        image.thumbnail((200, 200))
        max_score = 0
        dominant_color = 0
        for count, (r, g, b, a) in image.getcolors(image.size[0] * image.size[1]):
            # 跳过纯黑色
            if a == 0:
                continue
            saturation = colorsys.rgb_to_hsv(r / 255.0, g / 255.0, b / 255.0)[1]
            y = min(abs(r * 2104 + g * 4130 + b * 802 + 4096 + 131072) >> 13, 235)
            y = (y - 16.0) / (235 - 16)
            # 忽略高亮色
            if y > 0.9:
                continue
            # Calculate the score, preferring highly saturated colors.
            # Add 0.1 to the saturation so we don't completely ignore grayscale
            # colors by multiplying the count by zero, but still give them a low
            # weight.
            score = (saturation + 0.1) * count
            if score > max_score:
                max_score = score
                dominant_color = (r, g, b)
        return np.array(dominant_color)
    
    @staticmethod
    def process_batch_image(img_list,algorithm = 'centroid', show_plot=0, **dict):
        '''
        input: 
        output: img_list的聚类中心列表
        当algorithm = 'mult_centers'时， dict需要给到值 ，如n_cls = 12, m_cls = 5
        '''

        Center_batch_image = []
        for i,img in enumerate(img_list):
            try:
                feature = MajorTones.preProces(img)
            except:
                print(f'文件{img}处理出错，以跳过')
                Center_batch_image.append(np.array([2, 2, 2]))  # 添加一个(2,2,2)数组到Center中占位，以表示出错
                continue

            if len(feature) == 0: # 无效图片
                Center_batch_image.append(np.array([2,2,2])) # 添加一个(2,2,2)数组到Center中，以表示出错
                continue
            # ================================================================
            if algorithm == 'centroid':
                centroid = MajorTones.getCentroid(feature)
                Center_batch_image.append(centroid)
            elif algorithm == 'mult_centers':
                centers, percent_centers = MajorTones.getCenters(feature,dict['n_cls'],dict['m_cls'])
                Center_batch_image.append((centers,percent_centers)) # 列表中每个元素是一个元组
            elif algorithm == 'other':
                centroid = MajorTones.get_dominant_color(img) # other
                Center_batch_image.append(centroid)
            # ================================================================
            proces_percent = (i+1)/len(img_list) * 100
            print(f'\r{proces_percent:.1f}%已完成',end='')
            if show_plot:
                plt.subplot(1,2,1)
                h,w,c = feature.shape
                plt.imshow(feature.reshape(h,w,c))
                plt.subplot(1,2,2)
                if algorithm == 'mult_centers':
                    plt.imshow(centers.reshape(len(centers), 1, 3))
                else:
                    plt.imshow(centroid.reshape(1,1,3))
                plt.title(f'{i}')
                plt.show()
                print(60*'*')
        
        # print('processing complete')
        return Center_batch_image


import cv2
import matplotlib.pyplot as plt
import numpy as np

class Dist:
    
    '''input: vec1 , vec2 均为列向量'''
    
    @staticmethod
    def euclidean_dist(vec1, vec2):
        """欧氏距离:
        我们现实所说的直线距离"""
        if type(vec1)==list:
            vec1, vec2 = np.array(vec1), np.array(vec2)
        assert vec1.shape == vec2.shape 

        return np.sqrt((vec2-vec1).T @ (vec2-vec1))
    
    @staticmethod
    def manhattan_dist(vec1, vec2):
        """曼哈顿距离:
        城市距离"""
        if type(vec1)==list:
            vec1, vec2 = np.array(vec1), np.array(vec2)
        assert vec1.shape == vec2.shape 
        
        return sum(abs(vec1 - vec2))

def get_gradient(coord, img, delta = 2):
    '''coord: [x,y] numpy array, img: gray image, dtype: float, if not :img/255
    get image gradient at specified coordinates 
    有没有可能这里的delta应该等于R
    '''
    H, W= img.shape
    if coord[0] < W-delta and coord[1] < H-delta:
        dh = img[coord[1], coord[0]+delta] - img[coord[1], coord[0]]
        dw = img[coord[1]+delta, coord[0]] - img[coord[1], coord[0]]
        mag = np.sqrt(dw*dw + dh*dh)
        return dh/delta,dw/delta, mag
    else:
        return 0,0,0

def get_randomPoint(refImage):
    h,w = refImage.shape[:2]
    x,y = np.random.randint(0,h), np.random.randint(0,w)
    return x,y

def get_point_color(coord, image):
    '''image: bgra channel
        coord: w,h array
    return  bgr color array'''
    bgr = cv2.cvtColor(image, cv2.COLOR_BGRA2BGR)
    h,w = image.shape[:2]
    coord[0] = np.clip(coord[0], 0, h-1)
    coord[1] = np.clip(coord[1], 0, w-1)
    b,g,r = bgr[coord[0], coord[1], 0], bgr[coord[0], coord[1], 1], bgr[coord[0], coord[1], 2]
    return np.array([b, g, r])


def makeSplineStroke(x0, y0, R, refImage, canvas):
    '''refImage: bgra channel
    取落笔时的颜色，绘制一个stroke
    '''
    strokeColor = get_point_color([y0, x0], refImage)
    K = []
    K.append([x0, y0])
    [x,y] = [x0, y0]
    [lastDx, lastDy] = [1e-6,1e-6]
    
    minStrokeLength, maxStrokeLength = 10, 30
    gray_img = cv2.cvtColor(refImage,cv2.COLOR_BGR2GRAY)
    for i in range(maxStrokeLength):

        refColor = get_point_color([y,x], refImage)
#         diff_1 = Dist.euclidean_dist(refColor, get_point_color([y,x], canvas))
#         diff_2 = Dist.euclidean_dist(refColor, strokeColor)
        diff_1 = Dist.manhattan_dist(refColor, get_point_color([y,x], canvas))
        diff_2 = Dist.manhattan_dist(refColor, strokeColor)

        if i > minStrokeLength and (diff_1 < diff_2):
            return K
        
        gx,gy, gradientMag = get_gradient([x,y], gray_img/255)
        # detect vanishing gradient
#         print('gradientMag', gradientMag)
        if gradientMag==0:
            return K
        # compute a normal direction
        dx, dy = -gy, gx
        # if necessary, reverse direction
        if lastDx * dx +lastDy * dy < 0:
            dx, dy = -dx, dy
        # filter the stroke direction
        fc = 1.
        dx = fc*dx + (1-fc)*lastDx 
        dy = fc*dy + (1-fc)*lastDy
       
        dx, dy = dx/np.sqrt(dx*dx + dy*dy), dy/np.sqrt(dx*dx + dy*dy)
        x,y = int(x+R*dx), int(y+R*dy)
        
        lastDx, lastDy = dx, dy
        K.append([x, y])
        
    return K


from IPython import display

def test():
    refImage = cv2.imread('temp/nuannuan.png')

    img = cv2.blur(refImage, (10,10))
    canvas = np.zeros(refImage.shape).astype('uint8')
    R = refImage.shape[0]/70 # -----------------------------
    for i in range(1500):
        x0, y0 = get_randomPoint(img)
        K = makeSplineStroke(x0, y0,R,img, canvas)

        color = get_point_color([y0,x0], img)
        b,g,r = color[0],color[1],color[2]

        b,g,r = int(b), int(g), int(r)

        for coord in K:
            circle = cv2.circle(canvas,(coord[0], coord[1]), 12, (r,g,b),-1)
        
        if i%50==0:
            display.clear_output(wait=True)
            plt.figure(figsize=(6,6))
            plt.imshow(circle);plt.show()
            
    return circle


if __name__ == '__main__':

    circle = test()
    


from scipy import sin, pi, cos, real, arcsin, arccos, diff, interpolate
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
font = FontProperties(fname=r"c:\windows\fonts\simsun.ttc", size=10)  # 显示中文

def theta_s(theta_i, p, m):
    '''输入的theta 是度数不是弧度,得到的是偏转角'''
    theta_i = theta_i * pi / 180
    theta_r = arcsin(sin(theta_i) / m)
    ths = (2 * p * theta_r - 2 * theta_i - (p - 1) * pi)
    ths = real(ths)
    res = ths * 180 / pi
    return res

def sgn(x):
    if x < 0:
        return -1
    elif x == 0:
        return 0
    else:
        return 1


def trans2thetap(theta_s, p):
    ''' 将偏转角转换为散射角 ,单位是角度，不是弧度
        p的范围需要≥0
    '''
    k_p = (180 - theta_s) // (2 * 180)
    q = sgn(theta_s + 360 * k_p)
    if q == 0:
        return theta_s
    theta_p = (theta_s + 2 * 180 * k_p) / q
    return theta_p


def find_theta_i(ths, p, m, error=1e-12):
    flag = 0  # 判断是否进行两次求值

    if p == 0 or p == 1:  # 0阶
        theta_min, theta_max = 0, 90


    else:
        # p=2时， 当 1<m<2.5时， 散射角真实位置处于 -0 ~ -360°
        theta_e = arccos(np.sqrt((m ** 2 - 1) / (p ** 2 - 1))) * 180 / pi
        theta_end = trans2thetap(theta_s(90, p, m), p)
        # 如果小于两者中更小的角度，则返回空
        if ths < theta_end and ths < trans2thetap(theta_s(theta_e, p, m), p):
            return []
        #         print('极值点:入射角{:.2f},入射角对应散射角{:.2f},thetaend:{:.2f}'.format(theta_e,theta_s(theta_e,p,m),theta_end))
        # 那么单值情况便只会出现在thetaend ~ 180° 散射角区域
        # 所以可以设定 当 theta < thetaend ，返回空
        #         if ths < theta_end : # 这一句可以直接将上面一句的and 改为 or
        #             return []        # 但，我想把这句注释掉，我想看到多条光线

        #        y_peak = trans2thetap(theta_s(theta_e,p,m),p) # 二阶光线的最高点，实际确实散射角最小位置
        theta_min = 1e-10
        theta_max = theta_e  # 这句话是不是有问题,应该是下面这句?
        #         theta_max = arccos(np.sqrt((m**2-1)/(p**2-1))) * 180/pi
        flag = 1

    temp_min = trans2thetap(theta_s(theta_min, p, m), p) - ths
    temp_max = trans2thetap(theta_s(theta_max, p, m), p) - ths

    # 二分求解
    while abs(temp_max - temp_min) > error:

        theta_mean = (theta_min + theta_max) / 2
        temp_mean = trans2thetap(theta_s(theta_mean, p, m), p) - ths
        if temp_mean * temp_max < 0:
            theta_min = theta_mean
            temp_min = temp_mean

        else:
            theta_max = theta_mean
            temp_max = temp_mean

    res = [theta_mean]
    if flag == 1:

        theta_min = theta_e
        theta_max = 90

        temp_min = trans2thetap(theta_s(theta_min, p, m), p) - ths
        temp_max = trans2thetap(theta_s(theta_max, p, m), p) - ths
        if temp_max * temp_min > 0:
            return res
        # 二分求解
        while abs(temp_max - temp_min) > error:

            theta_mean = (theta_min + theta_max) / 2
            temp_mean = trans2thetap(theta_s(theta_mean, p, m), p) - ths
            if temp_mean * temp_max < 0:
                theta_min = theta_mean
                temp_min = temp_mean

            else:
                theta_max = theta_mean
                temp_max = temp_mean

        res.append(theta_mean)

    return res


def phase(theta_i, alpha, p, m):
    '''输入的theta 是角度不是弧度'''
    theta_i = np.array(theta_i)
    theta_i = theta_i * pi / 180
    if p != 0:
        theta_i[theta_i > real(arcsin(m))] = None

    #     costheta_r = np.sqrt(1 - np.sin(theta_i/m)**2)
    theta_r = real(arcsin(sin(theta_i) / m))  # 这里默认给出复数形式，取实数部分，

    return 2 * alpha * (cos(theta_i) - p * m * cos(theta_r))


from scipy import diff


def phase_diff(ths, alpha, p0, p1, m):
    '''计算相位差'''

    theta_i0 = find_theta_i(ths, p0, m)
    theta_i1 = find_theta_i(ths, p1, m)

    try:
        if theta_i1[1]:
            theta_i1 = [theta_i1[1]]  # 这里如果find_theta_i返回两个解,选择第一个或者第二个
            # 而从Mie解中发现,这里对应的应该是第二个解
    except:
        pass
    theta_i0, theta_i1

    phase0 = phase(theta_i0, alpha, p0, m)
    phase1 = phase(theta_i1, alpha, p1, m)
    return phase1 - phase0


####################################
#        p0+p1 or p0 + p2
####################################
def pd_output(p, m, alpha):
    N = 500
    PD = np.zeros(N)

    if p == 1:
        theta_end = trans2thetap(theta_s(90, p, m), p)
        theta = np.linspace(1e-10, theta_end - 1e-10, N)
        theta_range = theta_end
    else:
        ths_end = arccos(np.sqrt((m ** 2 - 1) / (p ** 2 - 1))) * 180 / pi  # 彩虹角位置
        theta_end = trans2thetap(theta_s(ths_end, p, m), p)
        theta = np.linspace(theta_end + 1e-10, 180, N)
        theta_range = 180 - theta_end

    print(theta_end)

    dtheta = theta_range / (N - 1)
    i = 0
    for ths in theta:
        # result = find_theta_i(ths, p, m, error=1e-12)
        pd = phase_diff(ths, alpha, 0, p, m)
        PD[i] = pd
        i += 1

    # 输出: 散射角,相位差, 散射角, 相位差频率
    return (theta, PD, theta[1:], abs(diff(PD)) / (2 * pi * dtheta))




def frequency(p, m, theta_begin, alpha, dthe=3):
    Nu = np.zeros(len(alpha))
    if p == 1:
        flag_p = 1
        fName = '单角度数据//m={:.2f},p={:d},theta={:.2f}'.format(m, p, theta_begin)
        theta_max = trans2thetap(theta_s(90, p, m), p)
        # ************************************************************************************
        #         print('p={},m={:.2f},theta={},theta_max={:.2f}'.format(p,m,theta_begin,theta_max))
        # ************************************************************************************
        if theta_begin > theta_max:
            flag_p = 0
            # *********************************************************
    #             print('该角度范围无对应{}阶光线'.format(p))
    # *********************************************************

    if p == 2:
        flag_p = 2
        fName = '单角度数据//m={:.2f},p={:d},theta={:.2f}'.format(m, p, theta_begin)
        theta_e = arccos(np.sqrt((m ** 2 - 1) / (p ** 2 - 1))) * 180 / pi
        theta_min = min(trans2thetap(theta_s(90, p, m), p), trans2thetap(theta_s(theta_e, p, m), p))
        print('p={},m={:.2f},theta={},theta_min={:.2f}'.format(p, m, theta_begin, theta_min))
        if theta_begin < theta_min:
            flag_p = 0
            # *********************************************************
    #             print('该角度范围无对应{}阶光线'.format(p))
    # *********************************************************

    if flag_p:
        for i_alpha in range(len(alpha)):
            pdiff = []

            # 注意！！！！这里如果直接求导得到频率必须是将散射角间隔分为恰好一度！
            M = 2

            ths = np.linspace(theta_begin, theta_begin + dthe, M)
            for _ in ths:
                pdiff.append(phase_diff(_, alpha[i_alpha], 0, p, m))

            pdiff = np.array(pdiff).reshape(1, M)  # 将列表转为数组
            pdd = abs(diff(pdiff))  # 求导，斜率取正

            Nu[i_alpha] = pdd[0][0] / (2 * pi * dthe / (M - 1))
        return Nu, fName


def get_k1k2(theta_1_begin, theta_2_begin, dthe1, dthe2, p1=1, p2=1):
    """i1"""
    temp_x, temp_y_1, temp_y_2, temp_y_3 = [], [], [], []
    # 将找到极值的精确角度区间给theta1 theta2
    # **********************************************************
    theta1, theta2 = theta_1_begin, theta_2_begin
    # **********************************************************
    for i in range(1, 100):
        for j in range(2):
            if j == 0:
                theta = theta1
                p = p1
                temp_m = 1.3 + i * 0.001
                Nu1, _ = frequency(p, temp_m, theta_begin=theta, alpha=[1], dthe=dthe1)
                temp_x.append(temp_m)
                temp_y_1.append(Nu1[0])  # 以浮点数形式存放,而不是array
            if j == 1:
                theta = theta2
                p = p2
                # print("p=",p,"m=",temp_m,"theta_begin=",theta,"alpha=", [1],"dthe=",dthe2)
                Nu2, _ = frequency(p, temp_m, theta_begin=theta, alpha=[1], dthe=dthe2)
                temp_y_2.append(Nu2[0])  # 以浮点数形式存放,而不是array
    k1 = interpolate.interp1d(temp_x, temp_y_1, kind='cubic')
    k2 = interpolate.interp1d(temp_x, temp_y_2, kind='cubic')
    return k1, k2


def predict(v1, v2, k1, k2, m_min=1.301, m_max=1.399):
    """输入以及计算好m区间和角度区间的插值函数k1, k2
    然后给出v1, v2计算alpha,m"""
    func = lambda x: v1 * k2(x) - v2 * k1(x)

    def dichotomy2(f, xleft, xright, error=1e-12):
        fxleft = f(xleft)
        fxright = f(xright)

        if fxleft * fxright > 0:
            print('当前区域无解,返回右端点函数值\n')
            return xright
        while abs(xright - xleft) > error:
            xmean = (xleft + xright) / 2
            fxmean = f(xmean)
            if fxleft * fxmean < 0:
                xright = xmean
                fxright = fxmean
            else:
                xleft = xmean
                fxleft = fxmean
        return (xright + xleft) / 2

    m = dichotomy2(func, m_min, m_max)  # 1.30 ~ 1.35即可
    alpha1 = v1 / k1(m)
    alpha2 = v2 / k2(m)

    return alpha1, alpha2, m


from guang.sci.fft import space2fre
import numpy as np
from numpy import pi,sin,cos,linspace
import plotly.express as px
import streamlit as st



def example_func(x_min, x_max):
    x = np.linspace(x_min, x_max, 1000)
    f1, f2, f3 = 0.1, 2, 8
    w1, w2, w3 = 2 * pi * f1, 2 * pi * f2, 2 * pi * f3
    y = 2 * sin(w1 * x) + 0.8 * sin(w2 * x)  # + 0.2 * sin(w3*x)
    return x, y

def randFunc(N, F, A):
    x = np.linspace(-3, 4, N)
    y = np.zeros(N)
    for f, a in zip(F, A):
        w = 2 * np.pi * f
        y += np.sin(w * x) * a
    return x, y
def showLatexFunc(F, A):
    y = f" $f(x)=$ "
    for f, a in zip(F, A):
        w = f"$2\pi\cdot{f}$"
        y += f"{a}$\cdot\sin$ ({w} $x$) + "#f"$\\sin({w}x) a$"
    st.write(f"{y[:-2]} \n    $x\\in[-3, 4]$")


def tutorial():
    global example_func
    x, y = example_func(-10, 10)
    fig = px.line(x=x, y=y)
    "原始函数："
    r"$2 \cdot sin(2\pi\cdot 0.1  x) + 0.8 * sin(2\pi\cdot 2 x)$"
    st.write(fig)

    fre_y = np.fft.fft(y)
    X = np.arange(len(fre_y))
    fig2 = px.line(x=X, y=np.abs(fre_y),
                   labels={'x':'_Frequency', 'y':'Amplitude'},
                   range_x=(-50, 1050))
    "直接对函数进行fft"
    fig2



    fre_y_shift = np.fft.fftshift(fre_y)
    fig3 = px.line(x=X, y=np.abs(fre_y_shift),
                   labels={'x':'_Frequency', 'y':'_Amplitude'},
                   range_x=(-50, 1050))
    "fftshift移动为中心对称："
    fig3
    r""" 
    这是我们还没有给它设置含坐标刻度。它的很坐标的意义是频率值，
    中心点的频率为0，最右侧的频率应该是这个频率空间的最大频率值（听上去像废话），
    左边是跟右边对称的。那么最重要的就是确定出这个频率空间的最大频率是多少。  
    而这个最大频率是由原始空间（坐标空间）x决定。频率=1/周期，最大频率由最小周期所确定，而
    最小周期自然是由空间的最小间隔（或者说采样间隔）决定。  
    我们的x一共有1000个值，
    分布-10 ~ 10之间，那么x的最小间隔$\delta x =\frac{x_{max}-x_{min}}{N-1}= \frac{20}{1000-1}$，
    那么想象下，2个振荡周期最少需要几个坐标点？
    """
    r"""
    思考完毕，两个周期最小需要5个点。 想象一个锯齿函数~   
    那么N个点可以承载最多$\frac{N-1}{2}$个周期. 那时一个周期的距离就是两倍最小间隔$2\delta x$ . 
    这时我们就有： 
    $$
    T = 2\delta x,\ \ 
    f_{max} = \frac{1}{T} = \frac{1}{2\delta x}=\frac{N-1}{2(x_{max}-x_{min})}
    $$
    同样，对于频率f而言，有：
    $$
    \delta f = \frac{2f_{max}}{N-1}
    $$
    """

    x_max = x.max()
    x_min = x.min()
    N = len(x)
    delta_x = (x_max - x_min)/(N-1)

    f_max = 1/(2*delta_x)
    delta_f = 2*f_max/(N-1)
    r"这里计算得到$f_{max}=\frac{1}{2\cdot\delta x}=$", f_max
    r"$\delta f=$",delta_f
    if N%2==0:
        fre_x = np.linspace(-f_max,f_max,N)-delta_f/2
    else:
        fre_x = np.linspace(-f_max, f_max, N)
    fig4 = px.line(x=fre_x,y=np.abs(fre_y_shift), labels={"x":"frequency"},
                   )
    fig4

    """ 
    上图可以看到在频率为 0附近 以及 2附近存在两个峰，而这两个频率就是一开始时我们所设置的f1 和 f2 ,
    分别为0.1 和 2 x轴的意义弄清楚了。而y轴，对应的当然是振幅啦，只是我们需要进行一个规范化的操作：
    2*abs(fre_y_shift)/N-1
    """

    amplitude_y = 2*np.abs(fre_y_shift)/(len(x)-1)
    fig5 = px.line(x=fre_x, y=amplitude_y, labels={"x":"frequency", "y":"amplitude"})
    fig5


def test_fft():
    option = st.sidebar.selectbox("共有几个频率呢", np.arange(10))
    max_fre = 50
    f = np.arange(option)
    A = np.arange(option)
    for i in range(option):
        f[i] = st.sidebar.slider(f"频率{i}", 0, max_fre, 1)
        A[i] = st.sidebar.slider(f"振幅{i}", 0, 20, 5)

    N = st.sidebar.slider(f"函数采样数", 10, 10000, 1000)
    x, y = randFunc(N, f, A)
    fig0 = px.line(x=x, y=y)
    st.write(fig0)
    showLatexFunc(f, A)
    fre, amp = space2fre(x, y)

    fig = px.line(x=fre[:,0], y=amp[:,0],
                  labels={"x": "Frequency", "y": "Amplitude"})
    st.write(fig)

if st.sidebar.checkbox("Fundamental教程"):
    tutorial()

"---"
test_fft()




from functools import reduce
import numpy as np
import math

PI = math.pi
TAU = 2*PI

ORIGIN = np.array((0., 0., 0.))
UP = np.array((0., 1., 0.))
DOWN = np.array((0., -1., 0.))
RIGHT = np.array((1., 0., 0.))
LEFT = np.array((-1., 0., 0.))
IN = np.array((0., 0., -1.))
OUT = np.array((0., 0., 1.))
X_AXIS = np.array((1., 0., 0.))
Y_AXIS = np.array((0., 1., 0.))
Z_AXIS = np.array((0., 0., 1.))


def get_norm(vec):
    return sum(vec**2)**0.5
    # return sum([x**2 for x in vec])**0.5


def quaternion_mult(q1, q2):
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    return np.array([
        w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2,
        w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2,
        w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2,
        w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2,
    ])


def quaternion_from_angle_axis(angle, axis):
    return np.append(
        np.cos(angle / 2),
        np.sin(angle / 2) * normalize(axis)
    )


def angle_axis_from_quaternion(quaternion):
    axis = normalize(
        quaternion[1:],
        fall_back=np.array([1, 0, 0])
    )
    angle = 2 * np.arccos(quaternion[0])
    if angle > TAU / 2:
        angle = TAU - angle
    return angle, axis


def quaternion_conjugate(quaternion):
    result = np.array(quaternion)
    result[1:] *= -1
    return result


def rotate_vector(vector, angle, axis=OUT):
    if len(vector) == 2:
        # Use complex numbers...because why not
        z = complex(*vector) * np.exp(complex(0, angle))
        return np.array([z.real, z.imag])
    elif len(vector) == 3:
        # Use quaternions...because why not
        quat = quaternion_from_angle_axis(angle, axis)
        quat_inv = quaternion_conjugate(quat)
        product = reduce(
            quaternion_mult,
            [quat, np.append(0, vector), quat_inv]
        )
        return product[1:]
    else:
        raise Exception("vector must be of dimension 2 or 3")


def thick_diagonal(dim, thickness=2):
    row_indices = np.arange(dim).repeat(dim).reshape((dim, dim))
    col_indices = np.transpose(row_indices)
    return (np.abs(row_indices - col_indices) < thickness).astype('uint8')


def rotation_matrix(angle, axis):
    """
    Rotation in R^3 about a specified axis of rotation.
    """
    about_z = rotation_about_z(angle)
    z_to_axis = z_to_vector(axis)
    axis_to_z = np.linalg.inv(z_to_axis)
    return reduce(np.dot, [z_to_axis, about_z, axis_to_z])


def rotation_about_z(angle):
    return [
        [np.cos(angle), -np.sin(angle), 0],
        [np.sin(angle), np.cos(angle), 0],
        [0, 0, 1]
    ]


def z_to_vector(vector):
    """
    Returns some matrix in SO(3) which takes the z-axis to the
    (normalized) vector provided as an argument
    """
    norm = get_norm(vector)
    if norm == 0:
        return np.identity(3)
    v = np.array(vector) / norm
    phi = np.arccos(v[2])
    if any(v[:2]):
        # projection of vector to unit circle
        axis_proj = v[:2] / get_norm(v[:2])
        theta = np.arccos(axis_proj[0])
        if axis_proj[1] < 0:
            theta = -theta
    else:
        theta = 0
    phi_down = np.array([
        [np.cos(phi), 0, np.sin(phi)],
        [0, 1, 0],
        [-np.sin(phi), 0, np.cos(phi)]
    ])
    return np.dot(rotation_about_z(theta), phi_down)


def angle_between(v1, v2):
    return np.arccos(np.dot(
        v1 / get_norm(v1),
        v2 / get_norm(v2)
    ))


def angle_of_vector(vector):
    """
    Returns polar coordinate theta when vector is project on xy plane
    """
    z = complex(*vector[:2])
    if z == 0:
        return 0
    return np.angle(complex(*vector[:2]))


def fdiv(a, b, zero_over_zero_value=None):
    if zero_over_zero_value is not None:
        out = np.full_like(a, zero_over_zero_value)
        where = np.logical_or(a != 0, b != 0)
    else:
        out = None
        where = True

    return np.true_divide(a, b, out=out, where=where)

def angle_between_vectors(v1, v2):
    """
    Returns the angle between two 3D vectors.
    This angle will always be btw 0 and pi
    """
    return np.arccos(fdiv(
        np.dot(v1, v2),
        get_norm(v1) * get_norm(v2)
    ))


def project_along_vector(point, vector):
    matrix = np.identity(3) - np.outer(vector, vector)
    return np.dot(point, matrix.T)


def normalize(vect, fall_back=None):
    norm = get_norm(vect)
    if norm > 0:
        return np.array(vect) / norm
    else:
        if fall_back is not None:
            return fall_back
        else:
            return np.zeros(len(vect))


def cross(v1, v2):
    return np.array([
        v1[1] * v2[2] - v1[2] * v2[1],
        v1[2] * v2[0] - v1[0] * v2[2],
        v1[0] * v2[1] - v1[1] * v2[0]
    ])


def get_unit_normal(v1, v2):
    return normalize(cross(v1, v2))



def compass_directions(n=4, start_vect=RIGHT):
    angle = TAU / n
    return np.array([
        rotate_vector(start_vect, k * angle)
        for k in range(n)
    ])


def complex_to_R3(complex_num):
    return np.array((complex_num.real, complex_num.imag, 0))


def R3_to_complex(point):
    return complex(*point[:2])


def complex_func_to_R3_func(complex_func):
    return lambda p: complex_to_R3(complex_func(R3_to_complex(p)))


def center_of_mass(points):
    points = [np.array(point).astype("float") for point in points]
    return sum(points) / len(points)


def midpoint(point1, point2):
    return center_of_mass([point1, point2])


def line_intersection(line1, line2):
    """
    return intersection point of two lines,
    each defined with a pair of vectors determining
    the end points
    """
    x_diff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    y_diff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(x_diff, y_diff)
    if div == 0:
        raise Exception("Lines do not intersect")
    d = (det(*line1), det(*line2))
    x = det(d, x_diff) / div
    y = det(d, y_diff) / div
    return np.array([x, y, 0])





# install minimal components
!pip install -q parallel_wavegan PyYaml unidecode ConfigArgparse g2p_en nltk
!git clone -q https://github.com/espnet/espnet.git
!cd espnet && git fetch && git checkout -b v.0.6.1 1e8b6ce88d57b53d1b60cbb3f306652468b0ab63


# download pretrained models
url_transformer = "https://drive.google.com/open?id=1Hh3iHmyMyiWxuljDPLXgDejTRj5Unv2f"
parallelwavegan_config = "https://drive.google.com/open?id=1eIgEuILooSJ7Mqo4LhILVVz6zdFCBBOV"
url_parallelwavegan = "https://drive.google.com/open?id=1fxfqVFhxAdZs4NLlysWKGxnPPbcLtjtM"
train_no_dev_units = "https://drive.google.com/open?id=1sKazNb43clsbpp4HsJpS7Mx0oyPCZIdc"
model_json = "https://drive.google.com/open?id=1D6yR5xVk0wgViu1WJkcDP8dc6H_oXpxH"
cidian = "https://drive.google.com/open?id=1rfgx40GrYjip65A1TDA7m-31Yb6SFUHt"

from guang.Utils.google import download
download(url_transformer)
download(url_parallelwavegan)
download(parallelwavegan_config)
download(train_no_dev_units)
download(model_json)
download(cidian)
print("\n sucessfully finished download.")

import os
# set path
dict_path = "/content/train_no_dev_units.txt"
model_path = "/content/model.last1.avg.best"
vocoder_path = "/content/checkpoint-500000steps.pkl"
vocoder_conf = "/content/parallel_wavegan.v1.yaml"

import sys
sys.path.append("espnet")
import torch
device = torch.device("cuda")

# define E2E-TTS model
from argparse import Namespace
from espnet.asr.asr_utils import get_model_conf
from espnet.asr.asr_utils import torch_load
from espnet.utils.dynamic_import import dynamic_import
idim, odim, train_args = get_model_conf(model_path)
model_class = dynamic_import(train_args.model_module)
model = model_class(idim, odim, train_args)
torch_load(model_path, model)
model = model.eval().to(device)
inference_args = Namespace(**{"threshold": 0.5, "minlenratio": 0.0, "maxlenratio": 10.0})

# define neural vocoder
import yaml
from parallel_wavegan.models import ParallelWaveGANGenerator
with open(vocoder_conf) as f:
    config = yaml.load(f, Loader=yaml.Loader)
vocoder = ParallelWaveGANGenerator(**config["generator_params"])
vocoder.load_state_dict(torch.load(vocoder_path, map_location="cpu")["model"]["generator"])
vocoder.remove_weight_norm()
vocoder = vocoder.eval().to(device)

from guang.Utils.jupyter import reload
from guang.Voice.txt2phone import py2ph
# define text frontend

with open(dict_path) as f:
    lines = f.readlines()

lines = [line.replace("\n", "").split(" ") for line in lines]
char_to_id = {c: int(i) for c, i in lines}

def frontend(text):
    """Clean text and then convert to id sequence."""
    idseq = []
    for i in py2ph(text):
        idseq += [char_to_id[i]]
    idseq += [idim - 1]  # <eos>
    return torch.LongTensor(idseq).view(-1).to(device)

import time
from IPython.display import display, Audio

input_text = input("pls input\n")

with torch.no_grad():
    start = time.time()
    x = frontend(input_text)
    c, _, _ = model.inference(x, inference_args)
    z = torch.randn(1, 1, c.size(0) * config["hop_size"]).to(device)
    c = torch.nn.ReplicationPad1d(
        config["generator_params"]["aux_context_window"])(c.unsqueeze(0).transpose(2, 1))
    y = vocoder(z, c).view(-1)
rtf = (time.time() - start) / (len(y) / config["sampling_rate"])
print(f"RTF = {rtf:5f}")

display(Audio(y.view(-1).cpu().numpy(), rate=config["sampling_rate"]))



import numpy as np
import cv2
import matplotlib.pyplot as plt

def rotate(image, angle, center=None, scale=1.0):
    """Rotate the image angle degrees.
    if the center is None, initialize it as the center of
    the image
    """
    (h, w) = image.shape[:2]

    if center is None:
        center = (w // 2, h // 2)

    M = cv2.getRotationMatrix2D(center, angle, scale)
    rotated = cv2.warpAffine(image, M, (w, h))
    return rotated

def rotate_bound(image, angle):
    """Keep the boundary and rotate angle degrees clockwise"""
    (h, w) = image.shape[:2]
    (cX, cY) = (w / 2, h / 2)

    M = cv2.getRotationMatrix2D((cX, cY), -angle, 1.0)
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])

    nW = int((h * sin) + (w * cos))
    nH = int((h * cos) + (w * sin))

    M[0, 2] += (nW / 2) - cX
    M[1, 2] += (nH / 2) - cY

    return cv2.warpAffine(image, M, (nW, nH))

def auto_canny(image, sigma=0.33):
    
    v = np.median(image)
    lower = int(max(0, (1.0 - sigma) * v))
    upper = int(min(255, (1.0 + sigma) * v))
    edged = cv2.Canny(image, lower, upper)

    return edged

def cvtBlackWhite(fileName):
    '''
    convert image's black color to white and meanwhile convert white to black.
    '''
    img = cv2.imread(fileName, 0)
    img = 255-img
    cv2.imwrite('convert_'+fileName, img)
    
def cvt2rgb(img,channel = 'bgr'):
    '''it can convert image channel BGR,BGRA,HLS,HSV to RGB'''
    if channel=='bgr' or channel=='BGR' or channel=='BGRA':
        print('bgr',img.shape)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    elif channel=='HLS' or channel=='hls':
        img = cv2.cvtColor(img, cv2.COLOR_HLS2RGB)
    elif channel=='HSV' or channel=='hsv':
        img = cv2.cvtColor(img, cv2.COLOR_HSV2RGB)
    else:
        return 'bad channel'
    return img
    
def myplot(img,channel='bgr'):
    ''' can show image channel BGR,BGRA,HLS,HSV.
    if RGB, set channel = 0'''
    if channel:
        img = cvt2rgb(img, channel)
    plt.figure(figsize=(7,7))
    plt.imshow(img)
    plt.show()

if __name__ == "__main__":
    pass






import matplotlib.pyplot as plt
import numpy as np

colors = plt.cycler(color=
            ['red', 'blue', '#9988DD',
             '#EECC55', '#88BB44', '#FFBBBB'])
plt.rc('axes', prop_cycle=colors)

def circle_points(R=10, N=20):
    theta = np.linspace(0, np.pi * 2, N)
    x = R * np.cos(theta)
    y = R * np.sin(theta)
    return x, y

def connect(*vertex):
    """connect vertex[0]、vertex[1]、vertex[2] ..."""
    x, y = [], []
    for idx, point in enumerate(vertex):
        x.append(point[0])
        y.append(point[1])
        if len(x) == 2:
            plt.plot(x, y)
            x.pop(0)
            y.pop(0)




def drawCircles(R=10, N=25):
    X, Y = circle_points(R, N)
    plt.figure(figsize=(10,10))
    for idx, (x, y) in enumerate(zip(X,Y)):
        tX, tY = np.delete(X, idx), np.delete(Y, idx)
        for tx, ty in zip(tX, tY):
            connect([x, y], [tx, ty])
    plt.axis("equal")
    plt.show()


def quadrate(edge_length=10, centre=[0, 0]):
    vertex1 = [centre[0] -  edge_length/2, centre[1] +  edge_length/2]
    vertex2 = [centre[0] +  edge_length/2, centre[1] +  edge_length/2]
    vertex3 = [centre[0] +  edge_length/2, centre[1] -  edge_length/2]
    vertex4 = [centre[0] -  edge_length/2, centre[1] -  edge_length/2]
    return vertex1, vertex2, vertex3, vertex4

def line_eq(point1, point2):
    """get line equation (2D)
    return:
        1. the map of x, i.e. f(x)
        2. the coefficient a, b, c of  ax + by = c 
    """
    x1, y1 = point1[0], point1[1]
    x2, y2 = point2[0], point2[1]
    
    if x1 == x2:
#         return (1e16, np.inf), (np.inf, x1, y1, y2)
        x1 += 1e-15
    
    k = (y2 - y1)/(x2 - x1)
    f = lambda x: k*(x - x1) + y1
    b = k*x1 + y1
    
    A = k
    B = -1
    C = k*x1 - y1
    return (k, b), (A, B, C)

def kb2abc(k, b):
    A = k
    B = -1
    C = -b
    return A, B, C

def rot_mat(theta):
    mat = np.array([
        [np.cos(theta),  np.sin(theta)],
        [-np.sin(theta), np.cos(theta)]
    ])
    return mat

def k_rot(theta, k):
    A = k
    B = -1
    theta = theta * np.pi/180
    res= rot_mat(theta)@np.array([[A, B]]).T
    a = np.float(res[[0]])
    b = np.float(res[[1]])
    k = -a/b
    return k

def solve_eq(eq1, eq2):

    A = np.array([[eq1[0], eq1[1]], [eq2[0], eq2[1]]])
    b = np.array([eq1[2], eq2[2]]).T
    res = np.linalg.solve(A, b)
    return res

def slove_b(k, x, y):
    b = y-k * x
    return b

def special_(N=180, theta=3):
    plt.figure(figsize=(10, 10))
    v1, v2, v3, v4 = quadrate()
    connect(v1, v2, v3, v4, v1)
    (k0, b0), (A0,B0,C0) = line_eq(v1, v2)
    v = [v1, v2, v3, v4, v1]
    p0 = v[0]
    for idx in range(N):
        p1 = v[idx+1]
        p2 = v[idx+2]
        (k1, b1), coeff = line_eq(p1, p2)
        rot_k0 = k_rot(theta, k0)
        b = slove_b(rot_k0, *p0)
        point = solve_eq(kb2abc(rot_k0, b), coeff)
        connect(p0, point)
        p0 = point
        v.append(point)
        k0, b0 = k1, b1
    plt.axis("equal")
    plt.show()


if __name__ == "__main__":
    drawCircles()
    special_(N=2000, theta=1)


import numpy as np
import os
import shutil
import sys
from collections import Counter
from functools import wraps


def path(string):
    platform = sys.platform.lower()
    if 'linux' in platform:
        return string.replace('\\','/')
    elif 'win' in platform:
        return string.replace('/','\\')
    else:
        return string


def broadcast(func):
    '''
        example:
        @broadcast
        def f(x):
            # A function that can map only a single element
            if x==1 or x==0:
                return x
            else:
                return f(x-1)+f(x-2)
            
        >> f([2,4,10])
        >> array([1, 3, 832040], dtype=object)
    '''
    @wraps(func)
    def wrap(*args, **kwargs):
        '''
        Takes an arbitrary Python function and returns a NumPy ufunc
        Can be used, for example, to add broadcasting to a built-in Python function
        return: only one out, 
        type:numpy object

        '''
        nin, nout = len(args)+len(kwargs), 1
        return np.frompyfunc(func,nin, nout)(*args, **kwargs)
    return wrap

# :Enables the dictionary to be dot operated
class _Dict_enhance(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self 

def dict_dotable(dic):
    '''
    : input: a dictionary
    : output: an enhanced dictionary
    Example:
        enhance_dic = dict_dotable(dic)
    then, you can operate an enhanced dictionary like this:
        enhance_dic.key1.key2. ...
    '''
    dic = _Dict_enhance(dic)
    for i in dic:
        if type(dic[i]) == dict:
            dic[i] = dict_dotable(dic[i])
    return dic

# define a constant like C language.
class Cons:
    '''
    `object.__setattr__(self, name, value)`
    this built-in function will called when assigning values to properties of the class
    
    `object.__dict__` holds all writable attributes in object, 
    key as variable name and value as variable value.
    '''
    def __setattr__(self, name,value):
        if hasattr(self,name):
            raise ValueError('Constant value can\'t be changed')
        else:
            self.__dict__[name] = value

def rm(path):
    '''remove path
    '''
    if os.path.exists(path):
        if os.path.isfile(path):
            os.remove(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)
        else:
            print(f'{path} is  illegal !')

def index_char(L=1000):
    '''
    Get the index of all characters.
    use chr() 
    '''
    index_token = {}
    token = []
    for i in range(L):
        character = chr(i)
        index_token[i] = character
        token.append(character)
    token_index=dict(zip(token, range(L))) # token_index[idx] equal to ord(idx)
    return index_token, token_index


    
def yaml_dump(filepath, data):
    from yaml import dump
    try:
        from yaml import CDumper as Dumper
    except ImportError:
        from yaml import Dumper
    with open(filepath, "w", encoding='utf-8') as fw:
        fw.write(dump(data, Dumper=Dumper))
        
def yaml_load(filepath):
    from yaml import load
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader
    with open(filepath, 'r', encoding="utf-8") as stream:
    #     stream = stream.read()
        content = load(stream, Loader=Loader)
    return content

def sort_count(lis):
    '''
    返回lis的由大到小排序好的元素列表
    Example:
    l = np.array([2,2,2,2,5,5,3,9,9])
    sort_count(l) : [(2, 4), (5, 2), (9, 2), (3, 1)]
    # return [2, 5, 9,3], [4, 2, 2, 1]
    '''
    a = Counter(lis)
    b = sorted(a.items(),key=lambda item:item[1],reverse=True)
    # idx, counts = [b[i][0] for i in range(len(b))], [b[i][1] for i in range(len(b))]
    return b






import numpy as np

def space2fre(x, y):
    """ Transform space domain (x, y) to the frequency domain (frequency, amplitude)
    """
    _fre_y = np.fft.fft(y)
    fre_y_shift = np.fft.fftshift(_fre_y)
    x_max = x.max()
    x_min = x.min()
    N = len(x)
    delta_x = (x_max - x_min) / (N - 1)
    f_max = 1 / (2 * delta_x)
    # delta_f = 2 * f_max / (N - 1)
    fre_x = np.linspace(-f_max, f_max, len(x))
    amplitude_y = 2 * np.abs(fre_y_shift) / (len(x) - 1)
    idx = np.argwhere(fre_x>=0)
    return fre_x[idx], amplitude_y[idx]





import fire
import os
from guang.Utils.toolsFunc import path
from guang.cv.video import resample, embedFrameInfo



def mie():
    origin_wd = os.getcwd()
    os.chdir(r"C:\Users\beidongjiedeguang\OneDrive\a_github\ScatteringLight")
    os.system("streamlit run app.py")
    os.chdir(origin_wd)

def lorenz():
    app_path = path(os.path.join(os.path.dirname(__file__), "app/compose/anim_demo.py"))
    os.system(f"streamlit run {app_path}")

def sawtooth():
    file_path = path(os.path.join(os.path.dirname(__file__), "app/compose/slider.py"))
    os.system(f"python {file_path}")

def geo():
    file_path = path(os.path.join(os.path.dirname(__file__), "geo/compose.py"))
    os.system(f"python {file_path}")

def fourier():
    app_path = os.path.join(os.path.dirname(__file__), "sci/fourier_app.py")
    os.system(f"streamlit run {app_path}")

def multi_cvt2wav(PATH1, PATH2, FORMAT='*',sr=16000, n_cpu=None):
    """
    :arg PATH1: Input folder path
    :arg PATH2: Output folder path
    :arg FORMAT: Input voice format,  default all files
    :arg sr: Sample rate, default:16000
    :arg n_cpu: Number of multiple processes
    """
    from guang.Voice.convert import multi_cvt2wav as cvt
    cvt(PATH1, PATH2, FORMAT, sr, n_cpu)

def upload(PAHT1="upload/", PATH2="/var/www/html/"):
    from syncnote import Cloud

    def download(remote_path: str, local_path: str,
                 config_path: str = r"C:\Users\beidongjiedeguang\OneDrive\a_github\myWebsite\config.yaml"):
        cloud = Cloud(config_path)
        cloud.remote_walk(remote_path)
        cloud.local_walk(local_path)
        cloud.download()
    def upload(local_path: str, remote_path: str,
               config_path: str = r"C:\Users\beidongjiedeguang\OneDrive\a_github\myWebsite\config.yaml"):
        cloud = Cloud(config_path)
        cloud.remote_walk(remote_path)
        cloud.local_walk(local_path)
        cloud.upload()
    upload(PAHT1, PATH2)

# def embedFrameInfo(InputPath, OutputPath):
#     """嵌入视频当前帧数"""
#
#     embedFrameInfo(InputPath, OutputPath)




def main():
    fire.Fire({
        'mie':mie,
        'lorenz':lorenz,
        "geo":geo,
        'sawtooth':sawtooth,
        'cvt2wav':multi_cvt2wav,
        'fourier':fourier,
        'embedFrameInfo':embedFrameInfo,
        'resample':resample,

    })


if __name__ == "__main__":
    main()





import cv2
import numpy as np
import matplotlib.pyplot as plt
from vispy.color import (get_colormap, get_colormaps, Colormap)

class Render:
    ''' pass '''
    def __init__(self):
        pass
#         self.load = False

    def load_brush(self,brush):
        '''
            brush: 4 channel array
        '''
        self.B, self.G, self.R, self.A = cv2.split(brush)
        self.a_trans = self.A

    def load_A_channel(self, A):
        self.A = A
        self.a_trans = self.A
    def show_img(self):

        plt.imshow(cv2.merge((self.R,self.G,self.B)))
        plt.show()
        
    def emboss(self, img, intensity):
        '''input: 
        	img: 4D BGRA
        	The intensity of emboss
        '''
        img = cv2.cvtColor(img,cv2.COLOR_BGRA2RGB)
        (H,L,S) = cv2.split(cv2.cvtColor(img,cv2.COLOR_RGB2HLS))

        # Normalisation：For HSV, Hue range is [0,179], Saturation range is [0,255] and Value range is [0,255].
        H,L,S = H/179., L/255., S/255.
        H,L,S = H, L, S
        # Apply the convolution kernel to channel L： 
        p, q, s,t = 0.,0., intensity ,0.  
        # p,q,s control the intensity of the relief in three directions
        # The value of matrix center controls the overall brightness, here sets 1.0
        kernel = np.array([[-q, -s,  t],
                           [-p, 1.0, p],
                           [-t,  s,  q]], dtype=np.float32)
#         kernel = np.array([[ 0, -1,-2],
#                            [ 1, 0, -1],
#                            [ 2,  1, 0]], dtype=np.float32)
    
        L = cv2.filter2D(L, -1, kernel)
        L = np.clip(L,0,1)
        H,L,S = H*179., L*255., S*255.

        H,L,S = H.astype('uint8'),L.astype('uint8'),S.astype('uint8')
        #-------------------------------------------------------------------
        # If in the previous step makes the brightness too large, please re-limit the brightness here
        # H,L,S = H, Gamma_trans(L, 255, 1.2), Gamma_trans(S, 255, 1) 
        # H,L,S = H, MaxMinNormal(L, 1, 150), MaxMinNormal(S, 0,255)
        #-------------------------------------------------------------------
        # Merge each channel
        res = cv2.merge((H,L,S))
        res = cv2.cvtColor(res, cv2.COLOR_HLS2RGB)
        return cv2.cvtColor(res, cv2.COLOR_RGB2BGRA)
    
    @classmethod
    def colorMapInterp(cls, colors, N):
        '''
        colors: a list of RGB color array/list/tuple
                For example: colors = ([1,0,0],[0,1,0],...) 
                or colors = ('r','g') 
        N: number of interpolation 
        return: rgb interpolation array. type: float [0,1]
        '''
        rm = Colormap(colors)
        rgb_interp = rm[np.linspace(0,1,N)].rgb #* 255
        # show colors
        # plt.imshow(rgb_interp.reshape(1,N,3).astype('uint8'))
        # plt.axis('off')
        return rgb_interp # float [0,1]
    @classmethod
    def ramp(cls, c1, c2, c3, c4, h0=None, w0=None):
        '''
        input:
            c: Color of four vertices
        return:
            2D color bgr map array. type: float [0,1]
        '''
        if h0 == None:
            h0,w0 = cls.A.shape
        # scale = 20
        # h, w =  h0//scale+1, w0//scale+1 # h, w can be replace others and then we interpolate the output image
        h, w = 20, 20
        rec1 = cls.colorMapInterp( (c1, c2), h)

        rec3 = cls.colorMapInterp( (c3, c4), h)
 
        
        color_mat = np.zeros((h,w,3))
        for i in range(h):
            color_mat[i,:,:] = cls.colorMapInterp((rec1[i,:], rec3[h-i-1,:]), w)
        # show it
        # plt.imshow(color_mat)
        color_mat[:,:,[0, 2]] = color_mat[:,:,[2, 0]] # rgb -> bgr
        
        return cv2.resize(color_mat,(w0, h0)) # float [0,1]
    
    @staticmethod
    def get_random_color(bgr, sigma):
        '''
        rgb: [b, g, r] float value
        sigma: [sigma_b, sigma_g, sigma_r] float range
        return rgb color
        '''
#         color = [np.random.normal(bgr[2], sigma[0]), np.random.normal(bgr[1], sigma[1]), np.random.normal(bgr[0], sigma[2])]
        color = [bgr[2] + sigma[0]*(np.random.rand()-1/2), bgr[1]+sigma[1]*np.random.rand(), bgr[0]*sigma[2]+np.random.rand()]
        return np.clip(color, 0, 1)
    
    @classmethod
    def gen_colormap(cls, h, w, c1, sigma = [0.5, 0.5, 0.1]):
        '''
        c1: [b, g, r] float value
        sigma: [sigma_r, sigma_g, sigma_b] float range
        '''
        c2 = Render.get_random_color(c1, sigma)
        c3 = Render.get_random_color(c1, sigma)
        c4 = Render.get_random_color(c1, sigma)
        color=Render.ramp(c1,c2,c3,c4, h, w)
        return color
    

    def get_fore_back(self, canvas, brush):
        '''
        Get foreground and background of canvas and brush
        '''
        # brush = cv2.bilateralFilter(brush,15, 35,35)
        H,W = canvas.shape[:2] # canvas shape
        h_brush, w_brush = brush.shape[:2]

        h_coef = h_brush/w_brush # keep the same shape of brush
        scale = 0.4 # scale coefficient for roi
        h,w = int(H*h_coef*scale), int(W*scale) # roi shape
        # Attention! cv2.resize(img, (w,h)) NOT (h, w)
        brush = cv2.resize(brush, (w,h),interpolation=cv2.INTER_AREA)
#         plt.figure(figsize=(12,12))
#         plt.subplot(231)
#         plt.imshow(cv2.cvtColor(brush, cv2.COLOR_BGR2RGB)); plt.title('brush(padding)')
        # padding
        brush = cv2.copyMakeBorder(brush,
                                  int(np.ceil((H-h)/2)),int(np.floor((H-h)/2)),
                                  int(np.ceil((W-w)/2)),int(np.floor((W-w)/2)),
                                  cv2.BORDER_CONSTANT,value=[255,255,255])
#         print(canvas.shape, brush.shape)
#         plt.figure(figsize=(12,12))
#         plt.subplot(231)
#         plt.imshow(cv2.cvtColor(brush, cv2.COLOR_BGR2RGB)); plt.title('brush(padding)')

        brush = cv2.bilateralFilter(brush,15, 35,35) # this step can be delete if input image is 'clean'.
        brushgray = cv2.cvtColor(brush,cv2.COLOR_BGR2GRAY)    
        ret, mask = cv2.threshold(brushgray, 250, 255, cv2.THRESH_BINARY)#ret is thresh（250）mask is Binary image

        mask_inv = cv2.bitwise_not(mask)
        #plt.subplot(232);plt.imshow(mask);plt.title('mask')
        #plt.subplot(233);plt.imshow(mask_inv);plt.title('mask_inv')

        canvas_bg = cv2.bitwise_and(canvas, brush, mask=mask)# 将mask去掉，可以看出 _and操作是取数值更低（更黑）的元素,而mask是将mask对应元素置于0
        brush_bg = cv2.bitwise_and(canvas, canvas, mask=mask_inv)
        #plt.subplot(234);plt.imshow(cv2.cvtColor(canvas_bg, cv2.COLOR_BGR2RGB));plt.title('canvas_brush_mask_and')
        #plt.subplot(235);plt.imshow(cv2.cvtColor(brush_bg, cv2.COLOR_BGR2RGB));plt.title('brush_brush_mask_inv_and')

        brush_fg = cv2.bitwise_and(brush,brush,mask = mask_inv)
        #plt.subplot(236);plt.imshow(cv2.cvtColor(brush_fg, cv2.COLOR_BGR2RGB));plt.title('brush_fg')

        return canvas_bg, brush_bg, brush_fg # dtype: uint8


    def rotate_origin_img(self, img, theta):
        '''
        theta: Counterclockwise Angle
        This function is going to rotate and pad the image to the right size
        '''
        h_origin, w_origin = img.shape[:2]

        # padding
        size = int(np.ceil(np.sqrt(w_origin**2+h_origin**2)))
        top = int(np.ceil((size-h_origin)/2))
        bottom = int(np.ceil((size-h_origin)/2))
        left = int(np.ceil((size-w_origin)/2))
        right = int(np.ceil((size-w_origin)/2))
        padding_img = cv2.copyMakeBorder(img,top,bottom,left,right,cv2.BORDER_CONSTANT,value=[0,0,0])

        rotate_origin = ((size-1)/2,(size-1)/2)
        M = cv2.getRotationMatrix2D(rotate_origin, theta, 1)
        rotated_img = cv2.warpAffine(padding_img, M, (size, size), flags =cv2.INTER_LANCZOS4)

        assert rotated_img.dtype == 'uint8'
        return rotated_img, dict([('h_origin',h_origin), ('w_origin',w_origin),( 'theta',theta), 
                                  ('top',top), ('bottom',bottom), ('left',left), ('right',right), 
                                  ('rotate_origin',rotate_origin)])
    
    def inv_rotate_padding_img(self, padding_img, dic):

        h, w = padding_img.shape[:2]
        M_inv = cv2.getRotationMatrix2D(dic['rotate_origin'], - dic['theta'], 1)
        inv_rot_padding_img = cv2.warpAffine(padding_img, M_inv, (h, w), flags =cv2.INTER_LANCZOS4)

        # clip
        inv_rot_img = inv_rot_padding_img[dic['top']-1 :-dic['bottom'], 
                                          dic['left']-1:-dic['right'], ] # if dimension>2, add ':'

        inv_rot_img = cv2.resize(inv_rot_img, (dic['w_origin'], dic['h_origin']), interpolation=cv2.INTER_LANCZOS4)

        assert inv_rot_img.shape[:2] == (dic['h_origin'], dic['w_origin'])
        assert inv_rot_img.dtype == 'uint8'
        return inv_rot_img 
    

    
    def transp_mat(self, size, alpha_min, alpha_max, rate):
        '''
        Args:
            size: rotated image shape
        give the transparent matrix.
        '''
        N = 50
        mat = np.ones((N,N))
        x = np.linspace(1-alpha_max,1-alpha_min,N)
        alpha = ((1-utils.Gamma_trans(x,rate))*(alpha_max - alpha_min) + alpha_min)
        for i in range(N):
            mat[i,:] = alpha[i] * mat[i,:]
        mat = cv2.resize(mat,(size))
        return mat
    
    def transparency_rotate(self,  theta=0, alpha_min=0.3, alpha_max=1, rate=1):
        '''
        Apply transparency to channel A.
        Args
            a: channel A
            
        '''
        a = self.A
        a,coord_dic = self.rotate_origin_img(a, theta)
        a_trans = a * self.transp_mat(a.shape, alpha_min, alpha_max, rate)
        a_trans = self.inv_rotate_padding_img(a_trans.astype('uint8'), coord_dic)
        self.a_trans = a_trans
        
    def transparency_love(self, love_mat):
        h,w = self.A.shape
        love_mat = cv2.resize(love_mat, (w,h)) * 255
        self.a_trans = love_mat.astype('uint8')
    
    def overlay_with_transparency(self, bgimg,  xmin = 0, ymin = 0):
        '''
        bgimg: a 4 channel image, use as background
        xmin, ymin: a corrdinate in bgimg. from where the fgimg will be put
        trans_percent: transparency of fgimg. [0.0,1.0]
        '''
        trans_percent = 1
        fgimg = cv2.merge((self.B, self.G, self.R, self.a_trans))

        #create roi and black mask
        roi = bgimg[ymin:ymin+fgimg.shape[0], xmin:xmin+fgimg.shape[1]]
        roi_black = np.zeros(fgimg.shape).astype(bgimg.dtype)

        #convert alpha to [0,1] and reshape it so that we can to math operation later
        alpha = np.expand_dims((self.a_trans.astype('float32')/255.),axis = -1)
        alpha_inv = 1. - alpha

        roi_bg_black = (roi * alpha_inv).astype(bgimg.dtype)
        roi_bg_trans = cv2.addWeighted(roi, 1-trans_percent, roi_bg_black, trans_percent,0)

        fg_black = (fgimg * alpha).astype(fgimg.dtype)
        fg_trans = cv2.addWeighted(roi_black, 1-trans_percent, fg_black, trans_percent,0)

        result = bgimg.copy()
        result[ymin:ymin+fgimg.shape[0],xmin:xmin+fgimg.shape[1]] = cv2.add(fg_trans, roi_bg_trans)

        return result
# -------------------------------------------------------------
    def over_layer(self, img1, img2, transp1=1, transp2=1):
        '''
        input: float bgra
            img1: foreground   img2: background
            transp1: foreground transparency
        output: float bgra
        '''
        b1,g1,r1,alpha1 = cv2.split(img1)
        b2,g2,r2,alpha2 = cv2.split(img2)
        # ----------
        alpha1 = transp1 * alpha1
        alpha2 = transp2 * alpha2
        # ----------
        alpha = alpha1 + alpha2*(1-alpha1)
        alpha = np.clip(alpha, 0, 1)

        def outRGB(c1, c2):
            c1[alpha==0]=np.nan
            c2[alpha==0]=np.nan
            c = (c1*alpha1 + c2*alpha2*(1-alpha1))/alpha
            return np.nan_to_num(c)
        b = outRGB(b1,b2)
        g = outRGB(g1,g2)
        r = outRGB(r1,r2)
        b,g,r = np.clip(b,0,1), np.clip(g,0,1),np.clip(r,0,1)
        # b,g,r,alpha = utils.append1D(b,g,r,alpha)
        # bgra = np.concatenate([b,g,r,alpha], axis=2)
        bgra = cv2.merge([b,g,r,alpha])
        return bgra
    
    def overlay(self,img1, img2, xmin=0, ymin=0, transp_brush=1, transp_canvas=1):
        '''input: 
                float img1: foreground  img2: background
           return: float overlay
        '''
        roi = img2[ymin:ymin+img1.shape[0], xmin:xmin+img1.shape[1]]
        roi = self.over_layer(img1,roi, transp1 = transp_brush, transp2 = transp_canvas)
        img2[ymin:ymin+img1.shape[0], xmin:xmin+img1.shape[1]] = roi
        return img2
# -------------------------------------------------------------
    @staticmethod
    def zeroAlpha2zeroRGB(img):# this is not used
        '''
        input: 
            img: bgra 4d array, float
        return: bgra 4d array, float
        '''
        b,g,r,a = cv2.split(img)
        r[a==0] =0
        g[a==0] =0
        b[a==0] =0
        
        r,g,b,a = utils.append1D(r,g,b,a)
        return np.concatenate([b,g,r,a], axis=2)
    
class utils:
    @staticmethod
    def append1D(*array_list):
        ''' For exemple: 
                before: array.shape (5,2)
                after: array.shape  (5,2,1)
            '''
        out = []
        for array in array_list:
            out.append(array.reshape(array.shape+(1, )))
        if len(out)==1:
            return out[0]
        else:
            return [i for i in out]
    
    @staticmethod
    def myplot(img,channel='bgr',figsize=6):
        ''' can show image channel(float/uint8) BGR,BGRA,HLS,HSV.
        if RGB, set channel = 0'''

        def cvt2rgb(img,channel = 'bgr'):
            '''it can convert image channel BGR,BGRA,HLS,HSV to RGB'''
            if channel=='bgr' or channel=='BGR' :
                print('bgr',img.shape)
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            elif channel=='bgra' or channel=='BGRA' :
                print('bgra',img.shape)
#                 b,g,r,a = cv2.split(img)
#                 r,g,b,a = utils.append1D(r,g,b,a)
#                 img = np.concatenate([r,g,b,a], axis=2)
#                 img = np.c_[rgb, np.reshape(img[:,:,-1],rgb.shape[:2] +(1,))]
                img[:,:,[2,0]] = img[:,:,[0,2]]
            elif channel=='HLS' or channel=='hls':
                img = cv2.cvtColor(img, cv2.COLOR_HLS2RGB)
                print('hls',img.shape)
            elif channel=='HSV' or channel=='hsv':
                img = cv2.cvtColor(img, cv2.COLOR_HSV2RGB)
                print('hsv',img.shape)
            else:
                return img
            return img
        
        if img.dtype==np.float or img.dtype==np.float32:
            img = np.uint8(img*255)
        if channel:
            img = cvt2rgb(img, channel)
        plt.figure(figsize=(figsize,figsize))
        plt.imshow(img)
        plt.show()
        
    @staticmethod
    def Gamma_trans(I, gamma):
        '''
        param:
            gamma: if your intersted region is too bright, set gamma > 1 decreasing contrast.
               and if your intersted region is too dark, set 1> gamma > 0 to increase contrast.
            I_max: is the maximun of the channel of I.
        return:
            the map of I
        '''
        I_max = I.max()
        fI = I/I_max
        out = np.power(fI, gamma)
        out = out*I_max
        return out
    
if __name__ == '__main__':
    def test_overlay():
        canvas = cv2.imread('pic/brush_blue.png', -1)
        brush = cv2.imread('pic/pika_roi.png', -1)
        render = Render()
        canvas_ = render.overlay(brush/255, canvas/255, xmin=0, ymin=0, transp_brush=0.7)
        utils.myplot(canvas_, 'bgra')
    test_overlay()
    def test_ramp():
        render = Render()
        res = render.ramp('r', 'g', 'b', 'g')

    
    # =================================================================
#     render.load_brush(brush)
#     render.transparency_rotate(theta=0, alpha_min=0.2, alpha_max=0.99,rate=2)

#     res = render.overlay_with_transparency(canvas)
#     plt.imshow(cv2.cvtColor(res,cv2.COLOR_BGRA2RGB));plt.show()
#     res = render.emboss(res, 1)
#     plt.imshow(cv2.cvtColor(res,cv2.COLOR_BGRA2RGB));plt.show()
    # =================================================================